/**
 * @fileoverview This rule enforces to use appropriate i18n keys instead of hardcoding content
 * @author Aravindh T
 */

module.exports={
    meta:{
        type: "problem",
        docs: {
            description: "This rule prevents hardcoding content and enforces to use appropriate i18n keys"
        }

    },
    create: function(context){
        const excludes = context.options[0]
        const methods = excludes.methods
        const props = excludes.props
        const objectKeys = excludes.objectKeys
        const allowedTags = excludes.allowedTags
        const keys = excludes.keys
        const commentsToIgnoreHardcodingCheck = excludes.commentsToIgnoreHardcodingCheck
        const stringsAllowedAsBlockCommentStart = excludes.stringsAllowedAsBlockCommentStart
        const stringsAllowedAsBlockCommentEnd = excludes.stringsAllowedAsBlockCommentEnd
        const dateFormatPatterns = excludes.dateFormatPatterns
        const dateFormatPatternsRegExp = dateFormatPatterns.map((pattern) => {
                return new RegExp(pattern)
            })

        const ERROR_MESSAGE = 'Do not hardcode content. Use I18N key instead'
        let validTagsLocationObject = null  
        let aliasNamesOfExcludedMethods = []
        let aliasNamesOfgetI18NValueMethod = []
    
        return{
            
            Literal: function(node){
                //regexp patterns 
                const pattern = /(\w\s)+\w+/
                const cssStylingPattern= /([0-9]+(px|%|em|rem|vh|vw)(\s|$))+(\w+(\s|$))*(#[a-zA-Z0-9]+)?/
                const sequenceOfNumbers = /([0-9]+\s)+/
                 
                const propsWhereStringsCanBeAllowed = props
                const objectKeysWhereStringsCanBeAllowed = objectKeys
                const methodsWhereStringsCanBeAllowed = methods
                
                //boolean variables for reporting errors
                let literalText = node.value
                let isHardCoded = false
                let isASequenceOfNumbers = false
                let isAStringInACondition = false
                let isAStringInAllowedAttribute = false
                let isI18NNotNeeded = false
                let isAStringInAllowedProp = false
                let isAStringInAllowedTag = false
                let isACSSProperty = false
                let isAPropertyName  = false
                let isAMemberExpression = false
                let isAValidKey = false
                let isValidKeyInsideValidMethod = false
                let isAStringInAllowedMethod = false
                let isAStringInsideIgnoredBlock = false
                let isAStringInAttributeOfAllowedTag = false
                let isAStringInNewExpression = false
                let isAStringInThrowStatement = false
                let isADateFormatPattern = false

                //ancestors of literal node
                let ancestorsOfLiteral = context.getAncestors();
                let immediateAncestorOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -1]
                let grandParentOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -2]
                let greatGrandParentOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -3]

                //variables for checking comments
                let correctAncestors = ancestorsOfLiteral.find(node => node.type === 'Program')
                let comments = correctAncestors.comments.filter(Line => commentsToIgnoreHardcodingCheck.includes(Line.value.trim().toLowerCase()))
                let blockCommentsStart = correctAncestors.comments.filter(Line => stringsAllowedAsBlockCommentStart.includes(Line.value.trim().toLowerCase()))
                let blockCommentsEnd = correctAncestors.comments.filter(Line => stringsAllowedAsBlockCommentEnd.includes(Line.value.trim().toLowerCase()))
                let ancestorWithComments = ancestorsOfLiteral.find(ancestor => ancestor.type === 'JSXElement' && ancestor.parent.type === 'ReturnStatement')
                let stringStartsAt = node.loc.start.line
                let isValidBlockCommentAdded = false
                
                isHardCoded = pattern.test(literalText)
                isASequenceOfNumbers = sequenceOfNumbers.test(literalText)
                //checking comments
                if(isHardCoded && comments){
                    isI18NNotNeeded = comments.some(node =>node.loc.start.line === stringStartsAt)

                    if(!isI18NNotNeeded && immediateAncestorOfLiteral.type === 'CallExpression'){
                        isI18NNotNeeded = comments.some(node => node.loc.start.line === immediateAncestorOfLiteral.loc.start.line
                                                                || (immediateAncestorOfLiteral.loc.start.line < node.loc.start.line && 
                                                                    node.loc.start.line < immediateAncestorOfLiteral.loc.end.line)
                                                                || node.loc.start.line === immediateAncestorOfLiteral.loc.end.line)
                    }
                    if(!isI18NNotNeeded && ancestorWithComments){
                        isI18NNotNeeded = comments.some(node => node.loc.start.line === ancestorWithComments.parent.loc.start.line
                                                        ||  node.loc.start.line === ancestorWithComments.parent.loc.end.line)
                    }

                }
                //checking block comments
                if(blockCommentsStart && blockCommentsEnd){
                    isValidBlockCommentAdded = blockCommentsStart.length === blockCommentsEnd.length ? true : false
                }
                
                if(isHardCoded && isValidBlockCommentAdded){
                    for(let i=0;i<blockCommentsStart.length;i++){
                        if(blockCommentsStart[i].loc.start.line <= stringStartsAt
                            &&stringStartsAt<= blockCommentsEnd[i].loc.start.line){
                                isAStringInsideIgnoredBlock = true
                            }
                        if(isAStringInsideIgnoredBlock){
                            break;
                        }
                    }
                }

                //checking if string is a part of an allowed tag
                if(isHardCoded && validTagsLocationObject){
                    if((validTagsLocationObject.loc.start.line <= stringStartsAt)
                        && (stringStartsAt <= validTagsLocationObject.loc.end.line)){
                            isAStringInAllowedTag = true
                        }
                }
                //checking if a string is in an attribute of allowed tags
                let allowedTagAncestor = ancestorsOfLiteral.find(ancestor => ancestor && ancestor.type === 'JSXOpeningElement')
                isAStringInAttributeOfAllowedTag = allowedTags.some(tag => allowedTagAncestor && tag === allowedTagAncestor.name.name)
                // if(grandParentOfLiteral && grandParentOfLiteral.type === 'JSXOpeningElement'){
                //     isAStringInAttributeOfAllowedTag = allowedTags.some(tag => tag === grandParentOfLiteral.name.name)
                // }

                //checking if string is part of a condition
                if(immediateAncestorOfLiteral.type === 'BinaryExpression' ||
                immediateAncestorOfLiteral.type === 'ConditionalExpression' ||
                immediateAncestorOfLiteral.type === 'SwitchCase'){
                    isAStringInACondition = true
                }

                //checking if string is an allowed key/property
                
                if(immediateAncestorOfLiteral.type === 'Property'){
                    isAStringInAllowedAttribute = objectKeysWhereStringsCanBeAllowed.some(props => props===immediateAncestorOfLiteral.key.name)
                }
                else if(grandParentOfLiteral.type === 'Property'){
                    isAStringInAllowedAttribute = objectKeysWhereStringsCanBeAllowed.some(objectKey => objectKey===grandParentOfLiteral.key.name)
                }
                
                //checking if string is in an allowed prop
                if(immediateAncestorOfLiteral.type === 'JSXAttribute'){
                    isAStringInAllowedProp = propsWhereStringsCanBeAllowed.some(props => props === immediateAncestorOfLiteral.name.name)
                }
                else if(grandParentOfLiteral.type === 'JSXAttribute'){
                    isAStringInAllowedProp = propsWhereStringsCanBeAllowed.some(prop => prop === grandParentOfLiteral.name.name)
                    
                }
                else if(immediateAncestorOfLiteral.type === 'LogicalExpression' 
                && grandParentOfLiteral.type === 'JSXExpressionContainer'
                && greatGrandParentOfLiteral.type === 'JSXAttribute'){
                    isAStringInAllowedProp = propsWhereStringsCanBeAllowed.some(prop => prop === greatGrandParentOfLiteral.name.name)
                }

                //checking for css styling pattern
                if(isHardCoded){
                    isACSSProperty = cssStylingPattern.test(literalText)
                }

                //checking if string is a date format pattern
                //isADateFormatPattern = dateFormatPatternsRegExp.some(pattern => pattern.test(literalText))
                isADateFormatPattern = dateFormatPatternsRegExp.some(pattern => pattern.test(literalText))

                //checking if string is in a throw statement
                if(isHardCoded && immediateAncestorOfLiteral.type === 'ThrowStatement'){
                    isAStringInThrowStatement = true
                }               

                //checking if the strings is a property names
                if(isHardCoded && immediateAncestorOfLiteral.type=== 'Property' 
                && immediateAncestorOfLiteral.key.type === 'Literal'){
                    isAPropertyName = true
                }
                
                //omitting strings that are MemberExpression
                if(isHardCoded && immediateAncestorOfLiteral.type === 'MemberExpression'&&
                immediateAncestorOfLiteral.property.type === 'Literal'){
                    isAMemberExpression = true
                }

                //checking if string is in new expression
                if(immediateAncestorOfLiteral.type === 'NewExpression'){
                    isAStringInNewExpression = true
                }

                //checking if the string is a valid key and is used inside getI18NValue method
                isAValidKey = keys.includes(literalText)
                if(isHardCoded && isAValidKey){
                    if(immediateAncestorOfLiteral.type === 'CallExpression' && immediateAncestorOfLiteral.callee){
                        if(immediateAncestorOfLiteral.callee.type === 'MemberExpression'){
                            if(immediateAncestorOfLiteral.callee.property.name === 'getI18NValue'){
                                isValidKeyInsideValidMethod = true
                            }
                            else if(!isValidKeyInsideValidMethod){
                                isValidKeyInsideValidMethod = aliasNamesOfgetI18NValueMethod.some(method => method === immediateAncestorOfLiteral.callee.property.name)
                            }
                        }
                        else if(immediateAncestorOfLiteral.callee.type === 'Identifier'){
                            if(immediateAncestorOfLiteral.callee.name === 'getI18NValue'){
                                isValidKeyInsideValidMethod = true
                            }
                            else if(!isValidKeyInsideValidMethod){
                                isValidKeyInsideValidMethod = aliasNamesOfgetI18NValueMethod.some(method => method === immediateAncestorOfLiteral.callee.name)
                            }
                        }
                    }
                }
                
                //checking if the string is inside allowed method 
                if(isHardCoded){
                    if(immediateAncestorOfLiteral.type === 'CallExpression'){
                        if(immediateAncestorOfLiteral.callee && immediateAncestorOfLiteral.callee.type ==='MemberExpression'){
                            isAStringInAllowedMethod = methodsWhereStringsCanBeAllowed.some(method => 
                                                                                                method === immediateAncestorOfLiteral.callee.property.name)
                            if(!isAStringInAllowedMethod){
                                isAStringInAllowedMethod = aliasNamesOfExcludedMethods.some(method => method === immediateAncestorOfLiteral.callee.property.name)
                            }
                        }
                        else if(immediateAncestorOfLiteral.callee && immediateAncestorOfLiteral.callee.type === 'Identifier'){
                            isAStringInAllowedMethod = methodsWhereStringsCanBeAllowed.some(method => 
                                                                                                method === immediateAncestorOfLiteral.callee.name)
                            if(!isAStringInAllowedMethod){
                                isAStringInAllowedMethod = aliasNamesOfExcludedMethods.some(method => method === immediateAncestorOfLiteral.callee.name)
                            }
                        }
                    }
                    else if(immediateAncestorOfLiteral.type === 'Property'
                            && grandParentOfLiteral.type === 'ObjectExpression'
                            && greatGrandParentOfLiteral.type === 'CallExpression'){
                                if(greatGrandParentOfLiteral.callee.type === 'Identifier'){
                                    isAStringInAllowedMethod = methodsWhereStringsCanBeAllowed.some(method => method === greatGrandParentOfLiteral.callee.name)
                                }
                                else if(greatGrandParentOfLiteral.callee.type === 'MemberExpression'){
                                    isAStringInAllowedMethod = methodsWhereStringsCanBeAllowed.some(method => 
                                        method === greatGrandParentOfLiteral.callee.property.name)
                                }
                                
                            }
                }
                //reporting the errors
                if(isHardCoded && !isAStringInACondition && !isAStringInAllowedAttribute && !isAStringInAllowedProp 
                    && !isAStringInAttributeOfAllowedTag && !isAStringInAllowedTag && !isACSSProperty && !isAStringInThrowStatement && !isASequenceOfNumbers
                    && !isAPropertyName && !isAMemberExpression && !isAStringInNewExpression  && !isValidKeyInsideValidMethod && !isADateFormatPattern
                    && !isAStringInAllowedMethod && !isAStringInsideIgnoredBlock && !isI18NNotNeeded){
                    context.report({
                        node:node,
                        message: ERROR_MESSAGE
                    })
                }
            },
            JSXElement: function(node){
                let ancestors = context.getAncestors(node)
                let validAncestors = ancestors.filter(ancestor => ancestor.type ==='JSXElement')
                let allowedTagsLocation = []
                allowedTagsLocation = validAncestors.filter(ancestor => allowedTags.includes(ancestor.openingElement.name.name))
                                
                validTagsLocationObject = allowedTagsLocation[0]
            },
            VariableDeclarator: function(node){

                if(node.init && node.init.type === 'MemberExpression'){
                    aliasNamesOfExcludedMethods = methods.includes(node.init.property.name) ? [...aliasNamesOfExcludedMethods,node.id.name]
                                                                                            : [...aliasNamesOfExcludedMethods]
                }
                if(node.init && node.init.type === 'MemberExpression' && node.init.property.name === 'getI18NValue'){
                    if(aliasNamesOfgetI18NValueMethod.indexOf(node.id.name) === -1){
                        aliasNamesOfgetI18NValueMethod = [...aliasNamesOfgetI18NValueMethod,node.id.name]
                    }                    
                }                
            }
        }
    }
}